/*---------------------------------------*/
/*  CME 6993, HW#2, 1st MPI program      */
/*  Vince Chiarito, chiariv@wes.army.mil */
/*  16 Feb 2000       (in progress)   
/*  Input: none 
 *  Output: Trapezoidal rule estimate of integral
 *
 * Algorithm:
 *       1. Each process calulates a subinterval
 *          of integration, (b-a)/p, where p is the 
 *          number of processors used, and the total
 *          integration covers the range from a to b.
 *
 *       2. Each process estimates the integral of f(x)
 *          over its interval with the trapezoidal rule.
 *
 *       3. Each process other than 0 sends its integration
 *          result to process 0
 *   
 *       4. Process 0 receives the corresponding integration
 *           estimates from the other individual processes,
 *          sums the results, and prints the result of the
 *          integrations summed.
 *          
 *  Further Note:
 * This program computes the integral of 
 * the function : dx/[sqrt( 1 - cos(x)*cos(x)/2]
 * over from  0 to PI/2
 *  This integral is a complete elliptical integral of the 
 * first kind (recognizing that over the
 * range of the limits the integral has the same value
 * as the classical form of these integrals in terms
 * sin(x) instead of cos(x) ) with a value equal to 
 * 1.854075 (Jeffrey, Alan 1995.  Handbook of Mathematical
 * Formulas and Integrals, Academic Press, p.227, Table 12.1)
 */
/*---------------------------------------*/

 #include "stdio.h"
 #include "stdlib.h"
 #include "math.h"
 #include "mpi.h"  /* this includes acces to mpi library calls*/
 
main(int argc, char** argv) {
    int         p;             /* No. of processors */
    int         my_rank;       /* Processor rank */
    int         source;        /* rank of sender       */
    int         dest = 0;      /* rank of receiver     */
    int         tag = 0;       /* tag for messages     */
    MPI_Status  status;        /* return status for    */
                               /* receive              */
    double PI = 3.141592654; /* define the constant PI */
    int N = 128;         /* No. of intervals for integration */
    double a = 0.0;      /* global lower integration limit */
    double b = PI/2;     /* global upper integration limit */
    double deltax;     /* width of interval: trapezoid base length*/
    double local_a;      /* left endpoint of my process 
                          *  (local lower limit)           */
    double local_b;      /* right endpoint of my process
                          *  (local upper limit)             */
    int local_n;        /* number of trapezoids for calulation */
                        /* of ea. process */
    double fleftx;     /* left value of f(x)  */
    double frightx;    /* right value of f(x) */
    double psumintegral; /* partial sum of integration by ea. process */
    double total;        /* total integral */
  
                           /* local integration */
    double T_integrate(double local_a,
                       double local_b,
                       int local_n,
                       double deltax);

                        /* no mpi calls before the next statement */
    MPI_Init(&argc, &argv); /* initiate MPI */

                        /* Find out process rank  */
    MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
   /* printf("my_rank = %d"); */

                        /* Find out number of processes */
    MPI_Comm_size(MPI_COMM_WORLD, &p);
 printf("No. of processors = %d\n", p); 

/*---------------------------------------------------------*/
    deltax = (b - a)/N;  /*  same deltax for all processes */
    local_n = N/p;       /*  same no. of trapezoids for ea. process */
/*---------------------------------------------------------*/
/* Determine length of ea. process                         */
/* integration interval: local_n*deltax                    */
/*---------------------------------------------------------*/
    local_a = a + my_rank*local_n*deltax;   /* process my_rank starts */

  printf("local_a for process %d = %f\n",my_rank, local_a);

    local_b = local_a + local_n*deltax;  /* process my_rank ends */

    printf("local_b for process %d = %f\n", my_rank, local_b);

  psumintegral= T_integrate(local_a, local_b, local_n, deltax);
  printf(" process %d :contribution = %f\n", my_rank, psumintegral);

/*---------------------------------------------------------*/
/* sum the integrals computed by ea. process               */
/*---------------------------------------------------------*/
 if (my_rank == 0)
    {
        total = psumintegral;
          for (source = 1; source < p; source++)
          {
             MPI_Recv(&psumintegral, 1, MPI_FLOAT, source, tag,
                 MPI_COMM_WORLD, &status);
             total = total + psumintegral;
           }
    } else
      { MPI_Send(&psumintegral,1,MPI_FLOAT, dest, tag,
                  MPI_COMM_WORLD);
       }

/* print the result */
     if (my_rank == 0)
     {
    printf("With N = %d trapezoids,\n", N);
    printf("and N/p = %d trapezoids per process, the estimate \n",N/p);
    printf(" of the integral from %f to %f = %f\n", a,b,total);
     }
    MPI_Finalize(); /* shut down MPI */
                    /* no mpi calls after here */
 } /* end of main */
/*---------------------------------------------------------*/
  
 double T_integrate(double local_a /* in */,
                    double local_b /* in */,
                    int    local_n /* in */,
                    double deltax  /* in */)
   {         
   double PI = 3.141592654, psumintegral, x;
   int i;

   double f(double x);  /* integrand */
   
      psumintegral = 0.5*( f(local_a) + f(local_b) );
      x = local_a;
      for (i = 1; i <= local_n-1; i++)
      {
          x = x + deltax;
          psumintegral = psumintegral + f(x);
      }
       psumintegral = psumintegral*deltax;  
       return psumintegral;
   } /* end of T_integrate */
/*---------------------------------------------------------*/

 double f(double x) /* integrand function */

  {
   double y;
   y = cos(x);
   return (1.0/sqrt(1.0 - 0.5*y*y));
  }
/* end of function */
/*---------------------------------------------------------*/
